package xyz.zhouxy.plusone.jdbc;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.stream.Stream;

import javax.annotation.Nonnull;

import org.springframework.jdbc.core.ResultSetExtractor;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.util.CollectionUtils;

import xyz.zhouxy.plusone.commons.util.NumberUtil;
import xyz.zhouxy.plusone.exception.DataOperationResultException;

public abstract class PlusoneJdbcDaoSupport {
    protected final NamedParameterJdbcTemplate jdbc;

    protected PlusoneJdbcDaoSupport(@Nonnull NamedParameterJdbcTemplate namedParameterJdbcTemplate) {
        this.jdbc = namedParameterJdbcTemplate;
    }

    protected final <T> T queryForObject(String sql, ResultSetExtractor<T> resultSetExtractor) {
        return this.jdbc.query(sql, resultSetExtractor);
    }

    protected final <T> T queryForObject(String sql, SqlParameterSource paramSource,
            ResultSetExtractor<T> resultSetExtractor) {
        return this.jdbc.query(sql, paramSource, resultSetExtractor);
    }

    protected final <T> T queryForObject(String sql, String paramName, Object value,
            ResultSetExtractor<T> resultSetExtractor) {
        return this.jdbc.query(sql, new MapSqlParameterSource(paramName, value), resultSetExtractor);
    }

    protected final <T> List<T> queryForList(String sql, RowMapper<T> rowMapper) {
        return this.jdbc.query(sql, rowMapper);
    }

    protected final <T> List<T> queryForList(String sql, SqlParameterSource parameterSource, RowMapper<T> rowMapper) {
        return this.jdbc.query(sql, parameterSource, rowMapper);
    }

    protected final <T> List<T> queryForList(String sql, String paramName, Object value, RowMapper<T> rowMapper) {
        return this.jdbc.query(sql, new MapSqlParameterSource(paramName, value), rowMapper);
    }

    protected final <T> Stream<T> queryForStream(String sql, SqlParameterSource parameterSource,
            RowMapper<T> rowMapper) {
        return this.jdbc.queryForStream(sql, parameterSource, rowMapper);
    }

    protected final <T> Stream<T> queryForStream(String sql, String paramName, Object value, RowMapper<T> rowMapper) {
        return this.jdbc.queryForStream(sql, new MapSqlParameterSource(paramName, value), rowMapper);
    }

    protected final <T> Stream<T> queryForStream(String sql, SqlParameterSource parameterSource, Class<T> elementType) {
        return this.jdbc.queryForList(sql, parameterSource, elementType).stream();
    }

    protected final <T> Stream<T> queryForStream(String sql, String paramName, Object value, Class<T> elementType) {
        return queryForStream(sql, new MapSqlParameterSource(paramName, value), elementType);
    }

    protected final boolean queryForBool(String sql, SqlParameterSource parameterSource) {
        Boolean result = this.jdbc.queryForObject(sql, parameterSource, Boolean.TYPE);
        return Boolean.TRUE.equals(result);
    }

    protected final boolean queryForBool(String sql, String paramName, Object value) {
        return queryForBool(sql, new MapSqlParameterSource(paramName, value));
    }

    protected final int update(String sql, SqlParameterSource parameterSource) {
        return this.jdbc.update(sql, parameterSource);
    }

    protected final int update(String sql, String paramName, Object value) {
        return update(sql, new MapSqlParameterSource(paramName, value));
    }

    protected final long batchUpdate(String sql, SqlParameterSource[] batchArgs) {
        int[] i = this.jdbc.batchUpdate(sql, batchArgs);
        return NumberUtil.sum(i);
    }

    protected final <T> long batchUpdate(String sql, Stream<T> c,
            @Nonnull Function<T, SqlParameterSource> paramSourceBuilder) {
        int[] i = this.jdbc.batchUpdate(sql, buildSqlParameterSourceArray(c, paramSourceBuilder));
        return NumberUtil.sum(i);
    }

    protected final <T> long batchUpdate(String sql, Collection<T> c,
            @Nonnull Function<T, SqlParameterSource> paramSourceBuilder) {
        int[] i = this.jdbc.batchUpdate(sql, buildSqlParameterSourceArray(c, paramSourceBuilder));
        return NumberUtil.sum(i);
    }

    protected static final <T> void assertResultEquals(T result, T expectedValue) {
        if (!Objects.equals(result, expectedValue)) {
            throw new DataOperationResultException();
        }
    }

    protected static final <T> void assertResultEquals(T result, T expectedValue, Function<T, String> errMsg) {
        if (!Objects.equals(result, expectedValue)) {
            throw new DataOperationResultException(errMsg.apply(result));
        }
    }

    protected static final <T> void assertResultEquals(T result, T expectedValue, String msgTemplate, Object... args) {
        if (!Objects.equals(result, expectedValue)) {
            throw new DataOperationResultException(String.format(msgTemplate, args));
        }
    }

    protected static final <T, E extends Throwable> void assertResultEqualsOrThrow(T result, T expectedValue,
            Function<T, E> e) throws E {
        if (!Objects.equals(result, expectedValue)) {
            throw e.apply(result);
        }
    }

    protected static final void assertUpdateOneRow(int result) {
        assertResultEquals(result, 1);
    }

    protected static final void assertUpdateOneRow(int result, Function<Integer, String> errMsg) {
        assertResultEquals(result, 1, errMsg);
    }

    protected static final void assertUpdateOneRow(int result, String msgTemplate, Object... args) {
        assertResultEquals(result, 1, msgTemplate, args);
    }

    protected static final <E extends Throwable> void assertUpdateOneRowOrThrow(int result, Function<Integer, E> e)
            throws E {
        assertResultEqualsOrThrow(result, 1, e);
    }

    protected static final <T> SqlParameterSource[] buildSqlParameterSourceArray(
            T[] c,
            @Nonnull Function<T, SqlParameterSource> paramSourceBuilder) {
        if (c == null || c.length == 0) {
            return new SqlParameterSource[] {};
        }
        return buildSqlParameterSourceArray(Arrays.stream(c), paramSourceBuilder);
    }

    protected static final <T> SqlParameterSource[] buildSqlParameterSourceArray(
            Collection<T> c,
            @Nonnull Function<T, SqlParameterSource> paramSourceBuilder) {
        if (CollectionUtils.isEmpty(c)) {
            return new SqlParameterSource[] {};
        }
        return buildSqlParameterSourceArray(c.stream(), paramSourceBuilder);
    }

    protected static final <T> SqlParameterSource[] buildSqlParameterSourceArray(
            Stream<T> stream,
            @Nonnull Function<T, SqlParameterSource> paramSourceBuilder) {
        Objects.requireNonNull(stream);
        Objects.requireNonNull(paramSourceBuilder);
        return stream.map(paramSourceBuilder).toArray(SqlParameterSource[]::new);
    }
}
