package xyz.zhouxy.plusone.oss;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Objects;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.csource.common.MyException;
import org.csource.common.NameValuePair;
import org.csource.fastdfs.FileInfo;
import org.csource.fastdfs.StorageClient;
import org.csource.fastdfs.StorageServer;
import org.csource.fastdfs.TrackerClient;
import org.csource.fastdfs.TrackerServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.io.Files;

import lombok.Getter;
import xyz.zhouxy.plusone.commons.annotation.StaticFactoryMethod;
import xyz.zhouxy.plusone.exception.SysException;

public class FastDFSUtil {

    private final TrackerServer trackerServer;
    private final StorageServer storageServer;

    private static final Logger logger = LoggerFactory.getLogger(FastDFSUtil.class);

    FastDFSUtil() throws IOException, MyException {
        TrackerClient trackerClient = new TrackerClient();
        this.trackerServer = trackerClient.getTrackerServer();
        this.storageServer = trackerClient.getStoreStorage(trackerServer);
    }

    /**
     * 上传文件到 Fast DFS，失败将抛异常
     *
     * @param file 文件信息
     * @return 2 elements string array if success:<br>
     *         <ul>
     *         <li>results[0]: the group name to store the file</li>
     *         </ul>
     *         <ul>
     *         <li>results[1]: the new created filename</li>
     *         </ul>
     *         return null if fail
     * @throws FastDFSException
     */
    public String[] upload(FastDFSFile file) throws FastDFSException {
        logger.info("File Name: {}, File Length: {}", file.getFileName(), file.getContent().length);

        NameValuePair[] metaList = new NameValuePair[1];
        metaList[0] = new NameValuePair("author", file.getAuthor());

        long startTime = System.currentTimeMillis();
        StorageClient storageClient = null;
        String[] uploadResults = null;
        try {
            storageClient = new StorageClient(this.trackerServer, this.storageServer);
            uploadResults = storageClient.upload_file(file.getContent(), file.getExt(), metaList);

            if (uploadResults == null) {
                throw new FastDFSException("upload file fail, error code: " + storageClient.getErrorCode());
            }
            logger.info("Upload file successfully!!! group_name: {}, remoteFileName: {}, time used: {} ms",
                    uploadResults[0], uploadResults[1], System.currentTimeMillis() - startTime);

        } catch (IOException e) {
            throw new FastDFSException("IO Exception when uploadind the file:" + file.getFileName(), e);
        } catch (MyException e) {
            throw new FastDFSException(e);
        }

        return uploadResults;
    }

    public FileInfo getFile(String groupName, String remoteFileName) throws FastDFSException {
        try {
            StorageClient storageClient = new StorageClient(this.trackerServer, this.storageServer);
            return storageClient.get_file_info(groupName, remoteFileName);
        } catch (IOException e) {
            throw new FastDFSException("IO Exception: Get File from Fast DFS failed", e);
        } catch (Exception e) {
            throw new FastDFSException("Non IO Exception: Get File from Fast DFS failed", e);
        }
    }

    public InputStream downFile(String groupName, String remoteFileName) throws FastDFSException {
        try {
            StorageClient storageClient = new StorageClient(this.trackerServer, this.storageServer);
            byte[] fileByte = storageClient.download_file(groupName, remoteFileName);
            InputStream ins = new ByteArrayInputStream(fileByte);
            return ins;
        } catch (IOException e) {
            throw new FastDFSException("IO Exception: Get File from Fast DFS failed", e);
        } catch (Exception e) {
            throw new FastDFSException("Non IO Exception: Get File from Fast DFS failed", e);
        }
    }

    public void deleteFile(String groupName, String remoteFileName) throws FastDFSException {
        StorageClient storageClient = new StorageClient(this.trackerServer, this.storageServer);
        try {
            int i = storageClient.delete_file(groupName, remoteFileName);
            if (i == 0) {
                logger.info("Delete file SUCCESSFULLY!!!");
            } else {
                throw new FastDFSException("Delete file failed, error code is: " + i);
            }
        } catch (IOException | MyException e) {
            throw new FastDFSException(e);
        }
    }

    public static final class FastDFSFile {
        @Getter
        private final String fileName;
        private final byte[] content;
        @Getter
        private final String ext;
        @Getter
        private final String md5;
        @Getter
        private final String author;

        private FastDFSFile(@Nonnull File file, @Nullable String author) throws IOException {
            this.fileName = file.getName();
            this.content = Files.toByteArray(file);
            this.ext = Files.getFileExtension(fileName);
            this.md5 = md5Hex(content);
            this.author = author;
        }

        private FastDFSFile(@Nonnull String fileName, @Nonnull byte[] content, @Nullable String author) {
            this.fileName = fileName;
            this.content = content;
            this.ext = Files.getFileExtension(fileName);
            this.md5 = md5Hex(content);
            this.author = author;
        }

        @Nonnull
        @StaticFactoryMethod(FastDFSFile.class)
        public static FastDFSFile of(@Nonnull File file) throws IOException {
            return new FastDFSFile(file, null);
        }

        @Nonnull
        @StaticFactoryMethod(FastDFSFile.class)
        public static FastDFSFile of(@Nonnull File file, @Nullable String author) throws IOException {
            return new FastDFSFile(file, author);
        }

        @Nonnull
        @StaticFactoryMethod(FastDFSFile.class)
        public static FastDFSFile of(@Nonnull String fileName, @Nonnull byte[] content) {
            return new FastDFSFile(fileName, content, null);
        }

        @Nonnull
        @StaticFactoryMethod(FastDFSFile.class)
        public static FastDFSFile of(@Nonnull String fileName, @Nonnull byte[] content, @Nullable String author) {
            return new FastDFSFile(fileName, content, author);
        }

        public byte[] getContent() {
            return Arrays.copyOf(content, content.length);
        }

        @Nonnull
        private static String md5Hex(byte[] data) {
            try {
                MessageDigest messageDigest = MessageDigest.getInstance("MD5");
                messageDigest.update(data);
                byte[] result = messageDigest.digest();
                var sha512Hex = new BigInteger(1, result).toString(16);
                return Objects.requireNonNull(sha512Hex);
            } catch (NoSuchAlgorithmException e) {
                throw new SysException(e);
            }
        }
    }
}
