package xyz.zhouxy.plusone.system.application.common.model;

import java.util.Map;
import java.util.Optional;
import java.util.OptionalInt;
import java.util.OptionalLong;
import java.util.OptionalDouble;
import java.util.concurrent.ConcurrentHashMap;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import xyz.zhouxy.plusone.exception.BizException;

public class AuthenticationInfo<T> {

    /** token 值 */
    private final String token;

    /** 此 token 对应的LoginId，未登录时不会构建此类的实例 */
    private final T loginId;

    /** 账号类型 */
    private final String loginType;

    /** 登录设备类型 */
    private final String deviceType;

    /** 自定义数据 */
    @Nonnull
    private final Map<String, Object> metadata;

    private AuthenticationInfo(String token, T loginId, String loginType, String deviceType,
            @Nonnull Map<String, Object> metadata) {
        this.token = token;
        this.loginId = loginId;
        this.loginType = loginType;
        this.deviceType = deviceType;
        this.metadata = metadata;
    }

    public String getToken() {
        return token;
    }

    public T getLoginId() {
        return loginId;
    }

    public String getLoginType() {
        return loginType;
    }

    public String getDeviceType() {
        return deviceType;
    }

    private Object getObjFromMetadata(String key) {
        if (this.metadata.containsKey(key)) {
            return this.metadata.get(key);
        }
        throw new BizException(String.format("不存在 key 为 \"%s\"的元数据。", key));
    }

    public OptionalDouble getMetadataAsDouble(String key) {
        Object valObj = getObjFromMetadata(key);
        if (valObj instanceof Double val) {
            return OptionalDouble.of(val.doubleValue());
        }
        return OptionalDouble.empty();
    }

    public OptionalLong getMetadataAsLong(String key) {
        Object valObj = getObjFromMetadata(key);
        if (valObj instanceof Long val) {
            return OptionalLong.of(val.longValue());
        }
        return OptionalLong.empty();
    }

    public OptionalInt getMetadataAsInt(String key) {
        Object valObj = getObjFromMetadata(key);
        if (valObj instanceof Integer val) {
            return OptionalInt.of(val.intValue());
        }
        return OptionalInt.empty();
    }

    public Optional<String> getMetadataAsString(String key) {
        Object valObj = getObjFromMetadata(key);
        if (valObj instanceof String val) {
            return Optional.of(val);
        }
        return Optional.empty();
    }

    @SuppressWarnings("unchecked")
    public <C> Optional<C> getMetadata(String key, Class<C> type) {
        Object valObj = getObjFromMetadata(key);
        if (valObj == null || !type.isAssignableFrom(valObj.getClass())) {
            return Optional.empty();
        }
        return Optional.of((C) valObj);
    }

    // ==========================================
    // builder

    public static <T> Builder<T> builder(String token, T loginId, String loginType, String deviceType) {
        return new Builder<>(token, loginId, loginType, deviceType);
    }

    public static final class Builder<T> {
        private final String token;
        private final T loginId;
        private final String loginType;
        private final String deviceType;

        @Nonnull
        private final Map<String, Object> meta = new ConcurrentHashMap<>();

        private Builder(String token, T loginId, String loginType, String deviceType) {
            this.token = token;
            this.loginId = loginId;
            this.loginType = loginType;
            this.deviceType = deviceType;
        }

        public <C> Builder<T> putMetadata(String key, @Nullable C value) {
            this.meta.put(key, value);
            return this;
        }

        public AuthenticationInfo<T> build() {
            return new AuthenticationInfo<>(token, loginId, loginType, deviceType, meta);
        }
    }
}
